#!/bin/bash

# Notes:
# - Must mount /proc in the chroot or $ORIGIN doesn't work
#   because _dl_get_origin fails and LD_ORIGIN_PATH won't
#   work because you ignore it in SETUID.

# set -x

# If USE_SYSTEM is 1 then the system libraries from a path
# rooted in SYSROOT are copied into the chroots for use in
# the test. Otherwise the built glibc in BUILD is used.
USE_SYSTEM=0
SYSROOT=
BUILD=/home/carlos/build/glibc
TEST_USER=1000
TEST_GROUP=1000

# Create a sysroot in $path using $build built glibc, or if 
# USE_SYSTEM is 1 then the system libraries.
function startup_sysroot {
    local path="$1"
    local build="$2"

    mkdir -p "$path/root"
    mkdir -p "$path/usr/bin"
    mkdir "$path/lib64"
    if [ $USE_SYSTEM -eq 0 ]; then
	cp "$build/libc.so" "$path/lib64/libc.so.6"
	cp "$build/elf/ld.so" "$path/lib64/ld-linux-x86-64.so.2"
	cp "$build/dlfcn/libdl.so" "$path/lib64/libdl.so.2"
    else
	cp "$SYSROOT/lib64/libc.so.6" "$path/lib64/libc.so.6"
	cp "$SYSROOT/lib64/ld-linux-x86-64.so.2" "$path/lib64/ld-linux-x86-64.so.2"
	cp "$SYSROOT/lib64/libdl.so.2" "$path/lib64/libdl.so.2"
    fi
    mkdir "$path/proc"
    sudo mount --bind /proc "$path/proc"
    touch "$path/started.stamp"
}

# Stop and shutdown the sysroot.
function shutdown_sysroot {
    local path="$1"
    sudo umount "$path/proc"
    rm "$path/started.stamp"
}

function run_in_sysroot {
    local path="$1"
    local cmd="$2"
    if [ ! -e "$path/started.stamp" ]; then
	echo "ERROR: sysroot $path is not started!"
	exit 1
    fi
    sudo chroot --userspec=$TEST_USER:$TEST_GROUP "$path" "$cmd"
}

# Build libprimary.so and libsecondary.so used for tests.
# Each library is the same SONAME but built with a different PASS/FAIL text
# depending on the need of the test.

# Build the primary library.
gcc -DVALUE=0 -DMESSAGE="\"PASS: Correct primary library.\\n\"" -shared -fPIC -Wl,-soname,libprimary.so -o libprimary-pass.so primary.c
gcc -DVALUE=1 -DMESSAGE="\"FAIL: Wrong primary library.\\n\"" -shared -fPIC -Wl,-soname,libprimary.so -o libprimary-fail.so primary.c
# Build the secondary test library.
gcc -DMESSAGE="\"PASS: Correct secondary library.\\n\"" -shared -fPIC -Wl,-soname,libsecondary.so -o libsecondary-pass.so secondary.c
gcc -DMESSAGE="\"FAIL: Wrong secondary library.\\n\"" -shared -fPIC -Wl,-soname,libsecondary.so -o libsecondary-fail.so secondary.c

# Run all tests twice with non-curly braced DST and curly-braced DST
# variants.

# Test 1:
# - Verify that $ORIGIN is scanned first and that the right version
#   of the library is loaded, not the ./ one.
function test1 {
    local rpath='$ORIGIN:./'
    local curly=$1
    local testdir="test1dir"

    if [ "$curly" = "curly" ]; then
	rpath='${ORIGIN}:./'
	testdir="test1dirC"
    fi

    rm -rf $testdir
    echo "Test 1 $curly: Test that RPATH of $rpath loads from \$ORIGIN first."
    startup_sysroot "./$testdir" "$BUILD"
    cp libprimary-pass.so ./$testdir/root/libprimary.so
    cp libprimary-fail.so ./$testdir/libprimary.so
    gcc -g3 -O0 -L./$testdir -lprimary -o test1 ./test.c -Wl,-rpath="$rpath"
    mv test1 ./$testdir/root/
    run_in_sysroot ./$testdir /root/test1
    shutdown_sysroot "./$testdir"
}

test1
test1 curly

# Test 1a:
# - Verify that $ORIGIN$ORIGIN is scanned first and that the right version
#   of the library is loaded, not the /lib64 one.
function test1a {
    local rpath='$ORIGIN$ORIGIN'
    local curly=$1
    local testdir="test1adir"

    if [ "$curly" = "curly" ]; then
	rpath='${ORIGIN}${ORIGIN}'
	testdir="test1adirC"
    fi

    rm -rf $testdir
    echo "Test 1a $curly: Test that RPATH of $rpath loads from \$ORIGIN\$ORIGIN first."
    startup_sysroot "./$testdir" "$BUILD"
    mkdir -p "./$testdir/root/root"
    cp libprimary-pass.so ./$testdir/root/root/libprimary.so
    cp libprimary-fail.so ./$testdir/lib64/libprimary.so
    gcc -g3 -O0 -L./$testdir/lib64 -lprimary -o test1a ./test.c -Wl,-rpath="$rpath"
    mv test1a ./$testdir/root/
    run_in_sysroot ./$testdir /root/test1a
    shutdown_sysroot "./$testdir"
}

test1a
test1a curly

# Test 1b:
# - Verify that $ORIGIN{$ORIGIN} is scanned first and that the right version
#   of the library is loaded, not the /lib64 one.
function test1b {
    local rpath='$ORIGIN{$ORIGIN}'
    local curly=$1
    local testdir="test1bdir"

    if [ "$curly" = "curly" ]; then
	rpath='${ORIGIN}{${ORIGIN}}'
	testdir="test1bdirC"
    fi

    rm -rf $testdir
    echo "Test 1b $curly: Test that RPATH of $rpath loads from \$ORIGIN{\$ORIGIN} first."
    startup_sysroot "./$testdir" "$BUILD"
    mkdir -p "./$testdir/root{/root}"
    cp libprimary-pass.so "./$testdir/root{/root}/libprimary.so"
    cp libprimary-fail.so ./$testdir/lib64/libprimary.so
    gcc -g3 -O0 -L./$testdir/lib64 -lprimary -o test1b ./test.c -Wl,-rpath="$rpath"
    mv test1b ./$testdir/root/
    run_in_sysroot ./$testdir /root/test1b
    shutdown_sysroot "./$testdir"
}

test1b
test1b curly

# Test 1c:
# - Verify that $XRIGIN is not treated as $ORIGIN.
function test1c {
    local rpath='$XRIGIN'
    local curly=$1
    local testdir="test1cdir"

    if [ "$curly" = "curly" ]; then
	rpath='${XRIGIN}'
	testdir="test1cdirC"
    fi

    rm -rf $testdir
    echo "Test 1c $curly: Test that RPATH of $rpath loads from /lib64 only with invalid DST."
    startup_sysroot "./$testdir" "$BUILD"
    cp libprimary-fail.so "./$testdir/root/libprimary.so"
    cp libprimary-pass.so ./$testdir/lib64/libprimary.so
    gcc -g3 -O0 -L./$testdir/lib64 -lprimary -o test1c ./test.c -Wl,-rpath="$rpath"
    mv test1c ./$testdir/root/
    run_in_sysroot ./$testdir /root/test1c
    shutdown_sysroot "./$testdir"
}

test1c
test1c curly

# Test 2:
# - Verify that ./ is searched first, not $ORIGIN.
function test2 {
    local rpath='./:$ORIGIN'
    local curly=$1
    local testdir="test2dir"

    if [ "$curly" = "curly" ]; then
	rpath='./:${ORIGIN}'
	testdir="test2dirC"
    fi

    rm -rf $testdir
    echo "Test 2 $curly: Test that RPATH of $rpath loads from ./ first."
    startup_sysroot "./$testdir" "$BUILD"
    cp libprimary-pass.so ./$testdir/libprimary.so
    cp libprimary-fail.so ./$testdir/root/libprimary.so
    gcc -g3 -O0 -L./$testdir -lprimary -o test2 ./test.c -Wl,-rpath="$rpath"
    mv test2 ./$testdir/root/
    run_in_sysroot ./$testdir /root/test2
    shutdown_sysroot "./$testdir"
}

test2
test2 curly

# Test 3:
# - Verify $ORIGIN/ is searched and that it works. If ./ is used we fail.
function test3 {
    local rpath='$ORIGIN/:./'
    local curly=$1
    local testdir="test3dir"

    if [ "$curly" = "curly" ]; then
	rpath='${ORIGIN}/:./'
	testdir="test3dirC"
    fi

    rm -rf $testdir
    echo "Test 3 $curly: Test that RPATH of $rpath loads from \$ORIGIN/ first (trailing slash)."
    startup_sysroot "./$testdir" "$BUILD"
    cp libprimary-pass.so ./$testdir/root/libprimary.so
    cp libprimary-fail.so ./$testdir/libprimary.so
    gcc -g3 -O0 -L./$testdir -lprimary -o test3 ./test.c -Wl,-rpath="$rpath"
    mv test3 ./$testdir/root/
    run_in_sysroot ./$testdir /root/test3
    shutdown_sysroot "./$testdir"
}

test3
test3 curly

# Test 4:
# - In this non-setuid version we check that $ORIGIN is used to load
#   the first library, and then that the second library is loaded from
#   ./, showing both paths are active. If only $ORIGIN was used then
#   the dlopen fails.
function test4 {
    local rpath='$ORIGIN:./'
    local curly=$1
    local testdir="test4dir"

    if [ "$curly" = "curly" ]; then
	rpath='${ORIGIN}:./'
	testdir="test4dirC"
    fi

    rm -rf $testdir
    echo "Test 4 $curly: Test that RPATH of $rpath loads from both in order."
    startup_sysroot "./$testdir" "$BUILD"
    cp libprimary-pass.so ./$testdir/root/libprimary.so
    cp libprimary-fail.so ./$testdir/libprimary.so
    # Place secondary only in second search path to show both paths active.
    cp libsecondary-pass.so ./$testdir/libsecondary.so
    gcc -g3 -O0 -L./$testdir -lprimary -lsecondary -o test4 ./test4.c -Wl,-rpath="$rpath"
    mv test4 ./$testdir/root/
    run_in_sysroot ./$testdir /root/test4
    shutdown_sysroot "./$testdir"
}

test4
test4 curly

# Test 5:
# - In setuid $ORIGIN is ignored, so only ./ works.
function test5 {
    local rpath='$ORIGIN:./'
    local curly=$1
    local testdir="test5dir"

    if [ "$curly" = "curly" ]; then
	rpath='${ORIGIN}:./'
	testdir="test5dirC"
    fi

    rm -rf $testdir
    echo "Test 5 $curly [SUID]: Test that RPATH of $rpath ignores \$ORIGIN and uses ./."
    startup_sysroot "./$testdir" "$BUILD"
    cp libprimary-pass.so ./$testdir/libprimary.so
    cp libprimary-fail.so ./$testdir/root/libprimary.so
    gcc -g3 -O0 -L./$testdir -lprimary -o test5 ./test.c -Wl,-rpath="$rpath"
    mv test5 ./$testdir/root/
    sudo chown root.root ./$testdir/root/test5
    sudo chmod u+s ./$testdir/root/test5
    run_in_sysroot ./$testdir /root/test5
    shutdown_sysroot "./$testdir"
}

test5
test5 curly

# Test 6:
# - Again, $ORIGIN is ignored, and only ./ works.
function test6 {
    local rpath='$ORIGIN:./'
    local curly=$1
    local testdir="test6dir"

    if [ "$curly" = "curly" ]; then
	rpath='${ORIGIN}:./'
	testdir="test6dirC"
    fi

    rm -rf $testdir
    echo "Test 6 $curly [SUID]: Test that RPATH of $rpath loads from ./ first."
    startup_sysroot "./$testdir" "$BUILD"
    cp libprimary-pass.so ./$testdir/libprimary.so
    cp libprimary-fail.so ./$testdir/root/libprimary.so
    gcc -g3 -O0 -L./$testdir -lprimary -o test6 ./test.c -Wl,-rpath="$rpath"
    mv test6 ./$testdir/root/
    sudo chown root.root ./$testdir/root/test6
    sudo chmod u+s ./$testdir/root/test6
    run_in_sysroot ./$testdir /root/test6
    shutdown_sysroot "./$testdir"
}

test6
test6 curly

# Test 7a:
# - Again, $ORIGIN/ (the whole path) is ignored, and only ./ works.
# - This test is sipmle, and we need 7b to test to make sure that $ORIGIN
#   isn't just being removed while leaving '/'.
function test7a {
    local rpath='$ORIGIN/:./'
    local curly=$1
    local testdir="test7dir"

    if [ "$curly" = "curly" ]; then
	rpath='${ORIGIN}/:./'
	testdir="test7dirC"
    fi

    rm -rf $testdir
    echo "Test 7a $curly [SUID]: Test that RPATH of $rpath ignores all \$ORIGIN paths and loads from ./."
    startup_sysroot "./$testdir" "$BUILD"
    cp libprimary-pass.so ./$testdir/libprimary.so
    cp libprimary-fail.so ./$testdir/root/libprimary.so
    gcc -g3 -O0 -L./$testdir -lprimary -o test7 ./test.c -Wl,-rpath="$rpath"
    mv test7 ./$testdir/root/
    sudo chown root.root ./$testdir/root/test7
    sudo chmod u+s ./$testdir/root/test7
    run_in_sysroot ./$testdir /root/test7
    shutdown_sysroot "./$testdir"
}

test7a
test7a curly

# Test 7b:
# - Verifies that if $ORIGIN is expaned to "" that the whole paths is also thrown away.
# - If it's not thrown away we are left with /root/ which would fail the test.
function test7b {
    local rpath='$ORIGIN/root/:./'
    local curly=$1
    local testdir="test7dir2"

    if [ "$curly" = "curly" ]; then
	rpath='${ORIGIN}/root:./'
	testdir="test7dir2C"
    fi

    rm -rf $testdir
    echo "Test 7b $curly [SUID]: Test that RPATH of $rpath ignores all \$ORIGIN paths and loads from ./."
    startup_sysroot "./$testdir" "$BUILD"
    cp libprimary-pass.so ./$testdir/libprimary.so
    cp libprimary-fail.so ./$testdir/root/libprimary.so
    gcc -g3 -O0 -L./$testdir -lprimary -o test7 ./test.c -Wl,-rpath="$rpath"
    mv test7 ./$testdir/root/
    sudo chown root.root ./$testdir/root/test7
    sudo chmod u+s ./$testdir/root/test7
    run_in_sysroot ./$testdir /root/test7
    shutdown_sysroot "./$testdir"
}

test7b
test7b curly

# Test 8:
# - Lastly, test that $ORIGIN is ignored and that dlopen uses ./ only.
function test8 {
    local rpath='$ORIGIN:./'
    local curly=$1
    local testdir="test8dir"

    if [ "$curly" = "curly" ]; then
	rpath='${ORIGIN}:./'
	testdir="test8dirC"
    fi

    rm -rf $testdir
    echo "Test 8 $curly [SUID]: Test that RPATH of $rpath loads only from ./."
    startup_sysroot "./$testdir" "$BUILD"
    cp libprimary-pass.so ./$testdir/libprimary.so
    cp libprimary-fail.so ./$testdir/root/libprimary.so
    cp libsecondary-pass.so ./$testdir/libsecondary.so
    cp libsecondary-fail.so ./$testdir/root/libsecondary.so
    gcc -g3 -O0 -L./$testdir -lprimary -o test8 ./test8.c -Wl,-rpath="$rpath" -ldl
    mv test8 ./$testdir/root/
    sudo chown root.root ./$testdir/root/test8
    sudo chmod u+s ./$testdir/root/test8
    run_in_sysroot ./$testdir /root/test8
    shutdown_sysroot "./$testdir"
}

test8
test8 curly

# Test 9:
# - Verify that $ORIGIN by itself does not work.
function test9 {
    local rpath='$ORIGIN'
    local curly=$1
    local testdir="test9dir"

    if [ "$curly" = "curly" ]; then
	rpath='${ORIGIN}'
	testdir="test9dirC"
    fi

    rm -rf $testdir
    echo "Test 9 $curly [SUID]: Test that RPATH of $rpath fails."
    startup_sysroot "./$testdir" "$BUILD"
    cp libsecondary-fail.so ./$testdir/root/libsecondary.so
    gcc -DMSGDSOFAIL="\"PASS: Failed to load library.\\n\"" \
    	-DMSGDSOLOAD="\"FAIL: Loaded library from \$ORIGIN.\\n\"" \
    	-g3 -O0 -o test9 ./test9.c -Wl,-rpath="$rpath" -ldl
    mv test9 ./$testdir/root/
    sudo chown root.root ./$testdir/root/test9
    sudo chmod u+s ./$testdir/root/test9
    run_in_sysroot ./$testdir /root/test9
    shutdown_sysroot "./$testdir"
}

test9
test9 curly

# Test 10:
# - Verify that $ORIGIN:$ORIGIN by itself does not work.
function test10 {
    local rpath='$ORIGIN:$ORIGIN'
    local curly=$1
    local testdir="test10dir"

    if [ "$curly" = "curly" ]; then
	rpath='${ORIGIN}:${ORIGIN}'
	testdir="test10dirC"
    fi

    rm -rf $testdir
    echo "Test 10 $curly [SUID]: Test that RPATH of $rpath fails."
    startup_sysroot "./$testdir" "$BUILD"
    cp libsecondary-fail.so ./$testdir/root/libsecondary.so
    gcc -DMSGDSOFAIL="\"PASS: Failed to load library.\\n\"" \
    	-DMSGDSOLOAD="\"FAIL: Loaded library from \$ORIGIN.\\n\"" \
    	-g3 -O0 -o test10 ./test9.c -Wl,-rpath="$rpath" -ldl
    mv test10 ./$testdir/root/
    sudo chown root.root ./$testdir/root/test10
    sudo chmod u+s ./$testdir/root/test10
    run_in_sysroot ./$testdir /root/test10
    shutdown_sysroot "./$testdir"
}

test10
test10 curly

# Test 11:
# - Verify that $ORIGIN$ORIGIN by itself does not work.
function test11 {
    local rpath='$ORIGIN$ORIGIN'
    local curly=$1
    local testdir="test11dir"

    if [ "$curly" = "curly" ]; then
	rpath='${ORIGIN}${ORIGIN}'
	testdir="test11dirC"
    fi

    rm -rf $testdir
    echo "Test 11 $curly [SUID]: Test that RPATH of $rpath fails."
    startup_sysroot "./$testdir" "$BUILD"
    mkdir ./$testdir/root/root
    # Place the library in $ORIGIN$ORIGIN e.g. /root/root/, which is
    # still not a trusted directory so the load should fail.
    cp libsecondary-fail.so ./$testdir/root/root/libsecondary.so
    gcc -DMSGDSOFAIL="\"PASS: Failed to load library.\\n\"" \
    	-DMSGDSOLOAD="\"FAIL: Loaded library from \$ORIGIN.\\n\"" \
    	-g3 -O0 -o test11 ./test9.c -Wl,-rpath="$rpath" -ldl
    mv test11 ./$testdir/root/
    sudo chown root.root ./$testdir/root/test11
    sudo chmod u+s ./$testdir/root/test11
    run_in_sysroot ./$testdir /root/test11
    shutdown_sysroot "./$testdir"
}

test11
test11 curly

# Test 12:
# - Verify that $ORIGIN/../../lib64/test12/ works, since it's canonical path
#   is rooted in the trusted /lib64.
# - Tested with $ORIGIN in a *trust* binary directory (shouldn't matter).
function test12 {
    local rpath='$ORIGIN/../../lib64/test12/'
    local curly=$1
    local testdir="test12dir"

    if [ "$curly" = "curly" ]; then
	rpath='${ORIGIN}/../../lib64/test12/'
	testdir="test12dirC"
    fi

    rm -rf $testdir
    echo "Test 12 $curly [SUID]: Test that RPATH of $rpath passes (trusted origin and path)."
    startup_sysroot "./$testdir" "$BUILD"
    mkdir ./$testdir/lib64/test12/
    cp libsecondary-pass.so ./$testdir/lib64/test12/libsecondary.so
    gcc -DMSGDSOFAIL="\"FAIL: Failed to load library.\\n\"" \
    	-DMSGDSOLOAD="\"PASS: Loaded library from path rooted in a trusted path.\\n\"" \
    	-g3 -O0 -o test12 ./test9.c -Wl,-rpath="$rpath" -ldl
    mv test12 ./$testdir/usr/bin/
    sudo chown root.root ./$testdir/usr/bin/test12
    sudo chmod u+s ./$testdir/usr/bin/test12
    run_in_sysroot ./$testdir /usr/bin/test12
    shutdown_sysroot "./$testdir"
}

test12
test12 curly

# Test 13:
# - Verify that $ORIGIN/../../lib64/test12/ works, since it's canonical path
#   is rooted in the trusted /lib64.
# - Tested with $ORIGIN in a non-trusted directory (shouldn't matter).
function test13 {
    local rpath='$ORIGIN/../../lib64/test13/'
    local curly=$1
    local testdir="test13dir"

    if [ "$curly" = "curly" ]; then
	rpath='${ORIGIN}/../../lib64/test13/'
	testdir="test13dirC"
    fi

    rm -rf $testdir
    echo "Test 13 $curly [SUID]: Test that RPATH of $rpath passes (trusted path only)."
    startup_sysroot "./$testdir" "$BUILD"
    mkdir -p ./$testdir/lib64/test13/
    cp libsecondary-pass.so ./$testdir/lib64/test13/libsecondary.so
    gcc -DMSGDSOFAIL="\"FAIL: Failed to load library.\\n\"" \
    	-DMSGDSOLOAD="\"PASS: Loaded library from path rooted in a trusted path.\\n\"" \
    	-g3 -O0 -o test13 ./test9.c -Wl,-rpath="$rpath" -ldl
    mv test13 ./$testdir/root/
    sudo chown root.root ./$testdir/root/test13
    sudo chmod u+s ./$testdir/root/test13
    run_in_sysroot ./$testdir /root/test13
    shutdown_sysroot "./$testdir"
}

test13
test13 curly

# Test 14:
# - Verify that $ORIGIN/../root/ fails, since it's canonical path is
#   not rooted in trusted directory (after canonicalization).
# - Tested with $ORIGIN in a trusted directory, to see if we can use
#   this to cause the implementation canonicalization to fail.
function test14 {
    local rpath='$ORIGIN/../root/'
    local curly=$1
    local testdir="test14dir"

    if [ "$curly" = "curly" ]; then
	rpath='${ORIGIN}/../root/'
	testdir="test14dirC"
    fi

    rm -rf $testdir
    echo "Test 14 $curly [SUID]: Test that RPATH of $rpath fails (untrusted path)."
    startup_sysroot "./$testdir" "$BUILD"
    cp libsecondary-fail.so ./$testdir/root/libsecondary.so
    gcc -DMSGDSOFAIL="\"PASS: Failed to load library.\\n\"" \
    	-DMSGDSOLOAD="\"FAIL: Loaded library from path rooted in a untrusted path.\\n\"" \
    	-g3 -O0 -o test14 ./test9.c -Wl,-rpath="$rpath" -ldl
    # Make sure $ORIGIN resolves to a trusted /lib64 path.
    mv test14 ./$testdir/lib64/
    sudo chown root.root ./$testdir/lib64/test14
    sudo chmod u+s ./$testdir/lib64/test14
    run_in_sysroot ./$testdir /lib64/test14
    shutdown_sysroot "./$testdir"
}

test14
test14 curly

# Test 15:
# - Verify that /$ORIGIN/liborigin.so in DT_NEEDED is removed in
#   AT_SECURE and the entire library element is ignored, rather
#   than just left uninterpreted as a directory named $ORIGIN.
#   We test for this by fabricating that false search path
#   e.g. /lib64/$ORIGIN/liborigin.so.
# 
function test15 {
    local soname='/$ORIGIN/liborigin.so'
    local curly=$1
    local testdir="test15dir"
    local ret

    if [ "$curly" = "curly" ]; then
	soname='/${ORIGIN}/liborigin.so'
	testdir="test15dirC"
    fi

    rm -rf $testdir
    echo "Test 15 $curly [SUID]: Verify \$ORIGIN with / in DT_NEEDED discards the DT_NEEDED."
    startup_sysroot "./$testdir" "$BUILD"
    gcc -shared -fPIC -Wl,-soname="$soname" -o liborigin.so origin.c
    gcc -g3 -O0 -L. -o test15 ./need_origin.c -lorigin
    mkdir -p "./$testdir/\$ORIGIN/"
    mkdir -p "./$testdir/lib64/\$ORIGIN/"
    cp liborigin.so "./$testdir/\$ORIGIN/"
    cp liborigin.so "./$testdir/lib64"
    cp liborigin.so "./$testdir/lib64/\$ORIGIN/"
    rm liborigin.so
    mv test15 ./$testdir/root/
    sudo chown root.root ./$testdir/root/test15
    sudo chmod u+s ./$testdir/root/test15
    run_in_sysroot ./$testdir /root/test15
    ret=$?
    if [ $ret -ne 0 ]; then
	echo "PASS: Failed to execute binary with \$ORIGIN in DT_NEEDED."
    else
	echo "FAIL: Incorrectly allowed DT_NEEDED with \$ORIGIN."
    fi
    shutdown_sysroot "./$testdir"
}

test15
test15 curly

# Test 16:
# - Verify that ${ORIGIN}liborigin.so in DT_NEEDED is removed in
#   AT_SECURE and the entire library element is ignored, rather
#   than just left uninterpreted as a directory named $ORIGIN.
#   We test for this by installing a library in /lib64 to match the
#   uninterpreted name.
# 
function test16 {
    local soname='/bad/${ORIGIN}liborigin.so'
    local testdir="test16dir"
    local ret

    rm -rf $testdir
    echo "Test 16 [SUID]: Verify \${ORIGIN} without / in DT_NEEDED discards the DT_NEEDED."
    startup_sysroot "./$testdir" "$BUILD"
    gcc -shared -fPIC -Wl,-soname="$soname" -o liborigin.so origin.c
    gcc -g3 -O0 -L. -o test16 ./need_origin.c -lorigin
    mkdir -p "./$testdir/bad"
    cp liborigin.so "./$testdir/bad/\${ORIGIN}liborigin.so"
    rm liborigin.so
    mv test16 ./$testdir/root/
    sudo chown root.root ./$testdir/root/test16
    sudo chmod u+s ./$testdir/root/test16
    run_in_sysroot ./$testdir /root/test16
    ret=$?
    if [ $ret -ne 0 ]; then
	echo "PASS: Failed to execute binary with \${ORIGIN} in DT_NEEDED."
    else
	echo "FAIL: Incorrectly allowed DT_NEEDED with \${ORIGIN}."
    fi
    shutdown_sysroot "./$testdir"
}

test16
# No curly version of test16.


echo Done.
exit 0
